/* FL2PP.C:  Far Lapp to PowerPak Adaptor

   The following code implements a few of the Phar Lapp DOS
   extender API functions using the PowerPak DPMI interrupts.
   The API functions implemented are:

            DosCreateCSAlias()
            DosCreateDSAlias()
            DosAllocRealSeg()
            DosFreeSeg()
            DosRealIntr()

   These are, of course, a small subset of the full Phar Lapp
   API interface but they should be sufficient to allow the
   PowerPak user to develop emulations for other Phar Lapp
   functions as needed.

*/

#include <dos.h>          // FP_SEG()
#include <mem.h>          // memset()
#include <stdarg.h>       // va_arg(), va_list(), va_start()
#include <alloc.h>        // malloc(), free()

#include "dpmi.h"         // DPMI stuff
#include "pharlap.h"      // Phar Lap stuff

/*--------------------------------------------------------------
   CreateAlias - returns either an executable selector (CS alias)
                 corresponding to a data selector  or a data
                 selector (DS alis) corresponding to a code
                 selector, depending on which of the values
                 passed in 'dataSel' and 'codeSel' are zero; if
                 'codeSel' is zero, 'dataSel' must be a valid
                 data selector and a corresponding CD alias will
                 be returned in 'codeSel' and vica versa. The
                 executable selector and data selector refer to
                 the same block of memory, but the executable
                 selector can be loaded into the CS segment
                 register, whereas in protected mode a data
                 selector cannot and the data selector can be
                 written to (poked) and read (peeked), whereas
                 in protected mode the code selector can only be
                 executed or read.

   ATTRIBS: standalone
   USAGE:
   RETURNS: On success, returns 0 and a valid CS/DS alias
            selector in the appropriate paramater (dataSel/
            codeSel) with the same base and limit as the
            data/code segment passed to the function. On failure
            returns DPMI error code.

   NOTES:   The selector created by this funtion should be
            released with DosFreeSeg() when it is no longer of
            use. This does not release the associated memory
            block until all selectors when reference the block
            are freed.

*/

USHORT APIENTRY CreateAlias(PSEL dataSel, PSEL codeSel) {
  BOOL code = FALSE;

  // One parameter must be zero and the other non-zero
  // otherwise we got'ta problem.
  if ((*codeSel && *dataSel) || (!*codeSel && !*dataSel))
    return DPMI_INVALIDVALUE;
  else
    if (*dataSel)
      code = TRUE;

  // Get alias for data/code selector and...
  if (code)
    _BX = *dataSel;
  else
    _BX = *codeSel;
  asm {
    mov ax, DPMI_CREATEALIASDESCRIP;
    int DPMINT;
    jc  error;
  }

  // ...plug it into appropriate location.
  if (code)
    *codeSel = _AX;
  else
    *dataSel = _AX;

/*
  // Now get alias descriptor's access rights word and...
  asm {
    mov bx, ax
    lar cx, bx
    jnz error
    shr cx, 8
  }

  // ...make this a code ('code' is non-zero) or data
  // ('code' is zero) selector.
  if (code)
    _CL |= 0x08;
  else
    _CL &= 0xf7;
  _AX = DPMI_SETDESCRIPACCESS;
  asm {
    int DPMINT
    jc  error
  }
*/
  return DPMI_SUCCESS;

error:
  return _AX;
} // end of CreateAlias()

/*--------------------------------------------------------------
   DosCreateCSAlias - returns an executable selector (CS alias)
                      corresponding to a data selector. The
                      executable selector and data selector
                      refer to the same block of memory, but the
                      executable selector can be loaded into the
                      CS segment register, whereas in protected
                      mode a data selector cannot.

   ATTRIBS: standalone
   USAGE:
   RETURNS: On success, returns 0 and a valid CS alias selector
            in 'pcodesel' with the same base and limit as the
            data segments passed to the function in 'datasel'.
            On failure returns DPMI error code.

   NOTES:   The selector created by this funtion should be
            released with DosFreeSeg() when it is no longer of
            use. This does not release the associated memory
            block until all selectors when reference the block
            are freed.

*/

USHORT APIENTRY DosCreateCSAlias(SEL datasel, PSEL pcodesel) {
  return CreateAlias(&datasel, pcodesel);
} // end of DosCreateCSAlias()


/*--------------------------------------------------------------
   DosCreateDSAlias - returns a data selector (DS alias) corres-
                      ponding to an executable selector. The
                      data and code selectors refer to the same
                      block of memory, but the data selector can
                      be written to (poked) and read (peeked),
                      whereas in protected mode the code selector
                      can only be executed.

   ATTRIBS: standalone
   USAGE:
   RETURNS: On success, returns 0 and a valid DS alias selector
            in 'pdatasel' with the same base and limit as the
            data segments passed to the function in 'codesel'.
            On failure returns DPMI error code.

   NOTES:   The selector created by this funtion should be
            released with DosFreeSeg() when it is no longer of
            use. This does not release the associated memory
            block until all selectors when reference the block
            are freed.

*/

USHORT APIENTRY DosCreateDSAlias(SEL codesel, PSEL pdatasel) {
  return CreateAlias(pdatasel, &codesel);
} // end of DosCreateDSAlias()


/*--------------------------------------------------------------
   DosAllocRealSeg - allocate a block of memory in conventional
                     memory. 'size' is the size of the block to
                     be allocated and CANNOT be larger than 64K.

   ATTRIBS: standalone

   USAGE:   PSEL   psel;
            USHORT ppara;
            if (DosAllocRealSeg(10000, &ppara, &psel)) {
              printf("\nCould not allocate 10,000 bytes of"
                     " conventional memory");
              exit(1);
            }
            DosFreeSeg(psel);

   RETURNS: On successful allocation (return value is 0) a
            segment is returned in 'ppara' and a selector
            pointer is returned in 'psel'. On failure a DPMI
            error code is returned.
*/

USHORT APIENTRY DosAllocRealSeg(ULONG size, PUSHORT ppara,
                                PSEL psel) {

  // DPMI_ALLOCREALSEG takes paragraphs so we need to convert
  // the bytes value passed in 'size' into paragraphs. Note that
  // this can lead to a DPMI_INVALIDVALUE error code being
  // returned even though the value passed in 'size' is less
  // than 65536 since we always round the paragraph value up
  // to the next highest paragraph value when the value passed
  // in 'size is not evenly divisible by 16.
  if (size%16)
    _BX = size+1;
  else
    _BX = size;

  if (size > 65536L)
    return DPMI_INVALIDVALUE;
    
  _AX = DPMI_ALLOCREALSEG;
  asm {
    int DPMINT;
    jc error;
  }

  *ppara = _AX;
  *psel  = _DX;
  return DPMI_SUCCESS;

error:
  return _AX;
} // end of DosAllocRealSeg()

/*--------------------------------------------------------------
   DosFreeSeg - releases a selector for a block of memory
                allocated via the DosAllocRealSeg() function.
                This function takes a selector and immediately
                frees the selector for reuse although the actual
                memory pointed to will not be released until all
                selectors pointing at the memory block are freed.

   RETURNS: On success returns 0 else returns DPMI error code.

*/

USHORT APIENTRY DosFreeSeg(USHORT sel) {
  _ES = 0;            //<--- make sure no selector in here!
  _DX = sel;
  _AX = DPMI_FREEREALSEG;
  asm {
    int DPMINT;
    jc error;
  }
  return DPMI_SUCCESS;

error:
  return _AX;
} // end of DosFreeSeg()

/*--------------------------------------------------------------
   DosRealIntr - generate real mode interrupt from protected
                 mode. Arguments may be passed in the REGS16
                 object or may be passed as arguments to this
                 function and are then pushed on the real mode
                 stack. Any interrupt data is returned in the
                 REGS16 object. If no 'words' are supplied in
                 the call to this routine, the real mode stack
                 is supplied by the DPMI server otherwise the
                 stack is supplied by this routine.

   RETURNS: On success returns 0 and data in 'pregs' (and/or on
            stack if 'words' is non-zero). On failure returns
            the non-zero DPMI error code.
*/

USHORT _far _cdecl DosRealIntr(USHORT intno, PREGS pregs,
                               ULONG reserved, USHORT words, ...) {

  RREGS       rr;
  va_list     va;
  int         arg;
  USHORT      our_stack;
  SEL         sel;
  USHORT      ret;
  USHORT _far *fp;
  int         i = 0;

  // If caller wants to use the stack...
  if (words != 0) {

    // ...get memory for stack from heap and...
    if ((ret = DosAllocRealSeg(words*sizeof(USHORT),
                               &our_stack, &sel)) != DPMI_SUCCESS)
      return PHARLAP_NOTENOUGHMEMORY;

    // ...load SS:SP with our stack...
    rr.ss = our_stack;
    rr.sp = FP_OFF(our_stack)+(words*sizeof(USHORT));
  }
  else

    // ...otherwise caller supplies stack or SS:SP is 0
    // and DPMI server supplies stack.
    rr.sp = pregs->sp;

  // Zero out the Borland DPMI register structure instance
  // and transfer the Phar Lap registers to it.
  memset(&rr, 0, sizeof(rr));
  if (pregs != 0) {
    rr.es = pregs->es;
    rr.ds = pregs->ds;
    rr.di = pregs->di;
    rr.si = pregs->si;
    rr.bp = pregs->bp;
    rr.bx = pregs->bx;
    rr.dx = pregs->bx;
    rr.cx = pregs->cx;
    rr.ax = pregs->ax;
    rr.ip = pregs->ip;
    rr.cs = pregs->cs;
    rr.flags = pregs->flags;
  }

  // Load Borland register instance, load caller's interrupt,
  // set stack words (if any) and...
  _ES = _SS;
  _DI = FP_OFF(&rr);
  _AX = DPMI_SIMULATEREALMODEINT;
  _BL = intno;
  _BH = 0;
  _CX = words;

  // ...if caller has words with us, stuff them.
  if (words) {
    va_start(va, words);
    while ((arg = va_arg(va, USHORT)) != 0) {
      fp = (PUSHORT)MAKEP(sel, i*(sizeof(USHORT)));
      *fp = (USHORT)arg;
    }
  }

  // Now do the Simulate-Real-Mode-Interrupt.
  asm int DPMINT;

  // Unless there was an error on that operation...
  ret = _AX;
  asm jc error;

  // ...transfer the results to the Phar Lap registers,...
  if (pregs != 0) {
    pregs->es    = rr.es;
    pregs->ds    = rr.ds;
    pregs->di    = rr.di;
    pregs->si    = rr.si;
    pregs->bp    = rr.bp;
    pregs->sp    = rr.sp;
    pregs->bx    = rr.bx;
    pregs->dx    = rr.bx;
    pregs->cx    = rr.cx;
    pregs->ax    = rr.ax;
    pregs->ip    = rr.ip;
    pregs->cs    = rr.cs;
    pregs->flags = rr.flags;
  }

  // ...free the stack (if any) and...
  if (words)
    DosFreeSeg(our_stack);

  // ...let caller know data is valid.
  return DPMI_SUCCESS;

  // Something went awry.
error:

  // Free the stack (if any) and...
  if (words)
    DosFreeSeg(our_stack);

  // ...let caller know data is junk.
  return ret;
} // end of DosRealIntr()


#if defined (FL2PP_MAIN)
#include <stdlib.h>
#include <stdio.h>

//--------------------------------------------------------------
VOID testCode(VOID) {
  asm {
    mov ax, '!'   // The '!' is 11 bytes from start of testcode
    int 0x29
  }
} // end of testCode()

//**************************************************************
int main(void) {
  REGS16 r;            // Phar Lap register struct instance
  SEL       sel;          // selector
  SEL       csel = 0;     // code selector
  SEL       dsel = 0;     // data selector
  USHORT    paras;        // segment
  ULONG     freeMemory;
  BYTE far  *memptr;

  if (!DosMemAvail(&freeMemory))
    printf("\nAvailable memory at start = %lu", freeMemory);

  // Test register structure parameter passing... this should
  // display an '!' using the undocumentated DOS fast display
  // interrupt 0x29.
  fail("\mDosRealIntr call failed",
       DosRealIntr(0x29, &r, 0L, 0));

  // Test use of real-mode stack parameter passing
  fail("\mDosRealIntr call failed",
       DosRealIntr(0x2f, &r, 0L, 3, 0x4546, 0x4748, 0x4950));

  // Test allocation of conventional memory block.
  fail("\nDosAllocRealSeg call failed",
       DosAllocRealSeg(1000, &paras, &sel));

  // Test free memory function by releasing the conventional
  // memory allocation made above.
  fail("\nDosFreeSeg call failed", DosFreeSeg(sel));

  return 0;
} // end of main()

#endif // FL2PP_MAIN

